"""Tests for Shape class"""
import os
import pytest

import vsdx
from vsdx import DataProperty
from vsdx import Page  # for typing
from vsdx import Shape
from vsdx import VisioFile

# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))


@pytest.mark.parametrize("filename, shape_id, expected_text", [
    ("test3_house.vsdx", "1", "Shape Text\n"),
    ("test3_house.vsdx", "11", "Shape to remove\n"),
    ("test2.vsdx", "9", "Group shape text\n"),
    ("test10_nested_shapes.vsdx", "5", "Shape 1.2.1\n"),
    ("test_master_multiple_child_shapes.vsdx", "3", "AWS Step Functions workflow \n")
    ])
def test_get_shape_text(filename: str, shape_id: str, expected_text: str):
    # Check that a specific shape on a page has expected text value
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_id(shape_id)
        # check that shape has expected text
        assert shape.text == expected_text


@pytest.mark.parametrize("filename, shape_id, expected_text", [
    ("test3_house.vsdx", "1", "Shape Text\n"),
    ("test3_house.vsdx", "11", "Shape to remove\n"),
    ("test2.vsdx", "9", "Group shape text\n"),
    ("test10_nested_shapes.vsdx", "5", "Shape 1.2.1\n"),
    ])
def test_set_shape_text(filename: str, shape_id: str, expected_text: str):
    # Check that a specific shape on a page has expected text value after it is updated
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_id(shape_id)
        # check that shape has expected text
        assert shape.text == expected_text
        shape.text = expected_text + 'changed'
        shape = page.find_shape_by_id(shape_id)
        assert shape.text == expected_text + 'changed'


@pytest.mark.parametrize("filename, attr, attr_value, expected_id", [
    ("test3_house.vsdx", "NameU", "House", "7"),
    ])
def test_get_shape_attr_value(filename: str, attr: str, attr_value: str, expected_id: str):
    # Check that a specific shape on a page has expected text value
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_attr(attr, attr_value)
        # check that shape has expected text
        assert shape.ID == expected_id


@pytest.mark.parametrize("filename, shape_id, child_count", [
    ("test3_house.vsdx", "7", 3),
    ("test3_house.vsdx", "11", 3),
    ("test2.vsdx", "9", 3),
    ("test10_nested_shapes.vsdx", "7", 2),
    ])
def test_get_shape_child_shapes(filename: str, shape_id: str, child_count: int):
    # Check that page has expected number of top level shapes
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_id(shape_id)

        # check that page has expected number of child shapes
        assert len(shape.child_shapes) == child_count


@pytest.mark.parametrize("filename, shape_id, all_count", [
    ("test3_house.vsdx", "7", 3),
    ("test3_house.vsdx", "11", 3),
    ("test2.vsdx", "9", 3),
    ("test10_nested_shapes.vsdx", "7", 6),
    ])
def test_get_shape_all_shapes(filename: str, shape_id: str, all_count: int):
    # Check that page has expected number of top level shapes
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page

        shape_group = page.find_shape_by_id(shape_id)
        for shape in shape_group.all_shapes:
            print(shape)
        # check that page has expected number of child shapes
        assert len(shape_group.all_shapes) == all_count


@pytest.mark.parametrize(
    "filename, expected_locations",
    [("test1.vsdx", "1.33,10.66 4.13,10.66 6.94,10.66 2.33,9.02 "),
     ("test2.vsdx", "2.33,8.72 1.33,10.66 4.13,10.66 5.91,8.72 1.61,8.58 3.25,8.65 ")])
def test_shape_locations(filename: str, expected_locations: str):
    print("=== list_shape_locations ===")
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shapes = page.child_shapes
        locations = ""
        for shape in shapes:  # type: Shape
            locations += f"{shape.x:.2f},{shape.y:.2f} "
        print(f"Expected:{expected_locations}")
        print(f"  Actual:{locations}")
    assert locations == expected_locations


@pytest.mark.parametrize(
    "filename, shape_id, expected_center",
    [
        ("test1.vsdx", "1", (1.332677148526936, 10.65551182326173)),
        ("test2.vsdx", "2", (1.082677148526936, 0.7874015625650443)),  # center of a group shape
        ("test2.vsdx", "16", (1.6903102768832179, 8.188976116607332)),  # test center of a line
    ])
def test_shape_center(filename: str, shape_id: str, expected_center: str):

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_id(shape_id)

        assert shape.center_x_y == expected_center


@pytest.mark.parametrize("filename", ["test2.vsdx", "test3_house.vsdx"])
def test_remove_shape(filename: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_shape_removed.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        # get shape to remove
        shape = vis.pages[0].find_shape_by_text('Shape to remove')  # type: Shape
        assert shape  # check shape found
        shape.remove()
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        # get shape that should have been removed
        shape = vis.pages[0].find_shape_by_text('Shape to remove')  # type: Shape
        assert shape is None  # check shape not found


@pytest.mark.parametrize(("filename", "shape_names", "shape_locations"),
                         [("test1.vsdx", {"Shape to remove"}, {(1.0, 1.0)})])
def test_set_shape_location(filename: str, shape_names: set, shape_locations: set):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_set_shape_location.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        # move shapes in list
        for (shape_name, x_y) in zip(shape_names, shape_locations):
            shape = vis.pages[0].find_shape_by_text(shape_name)  # type: Shape
            assert shape  # check shape found
            assert shape.x
            assert shape.y
            print(f"Moving shape '{shape_name}' from {shape.x}, {shape.y} to {x_y}")
            shape.x = x_y[0]
            shape.y = x_y[1]
        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        # get each shape that should have been moved
        for (shape_name, x_y) in zip(shape_names, shape_locations):
            shape = vis.pages[0].find_shape_by_text(shape_name)  # type: Shape
            assert shape  # check shape found
            assert shape.x == x_y[0]
            assert shape.y == x_y[1]


@pytest.mark.parametrize(("filename", "page_index", "shape_names", "shape_x_y_deltas"),
                         [("test1.vsdx", 0, {"Shape to remove"}, {(1.0, 1.0)}),
                          ("test4_connectors.vsdx", 0, {"Shape B"}, {(1.0, 1.0)}),
                          ("test4_connectors.vsdx", 1, {"B to C"}, {(1.0, 1.0)})])
def test_move_shape(filename: str, page_index: int, shape_names: set, shape_x_y_deltas: set):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_move_shape.vsdx')
    expected_shape_locations = {}

    with VisioFile(os.path.join(basedir, filename)) as vis:
        # move shapes in list
        for (shape_name, x_y) in zip(shape_names, shape_x_y_deltas):
            shape = vis.pages[page_index].find_shape_by_text(shape_name)  # type: Shape
            assert shape  # check shape found
            assert shape.x
            assert shape.y
            expected_shape_locations[shape_name] = (shape.x + x_y[0], shape.y + x_y[1])
            shape.move(x_y[0], x_y[1])

        vis.save_vsdx(out_file)
    print(f"expected_shape_locations={expected_shape_locations}")

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        # get each shape that should have been moved
        for shape_name in shape_names:
            shape = vis.pages[page_index].find_shape_by_text(shape_name)  # type: Shape
            assert shape  # check shape found
            assert shape.x == expected_shape_locations[shape_name][0]
            assert shape.y == expected_shape_locations[shape_name][1]


@pytest.mark.parametrize(("filename", "shape_name"),
                         [("test1.vsdx", "Shape to copy"),
                          ("test4_connectors.vsdx", "Shape B")])
def test_shape_copy(filename: str, shape_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_shape_copy.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page =  vis.pages[0]  # type: Page
        # find and copy shape by name
        shape = page.find_shape_by_text(shape_name)  # type: Shape
        assert shape  # check shape found
        print(f"found {shape.ID}")
        max_id = page.max_id

        new_shape = shape.copy()
        assert new_shape  # check new shape exists

        print(f"original shape {type(shape)} {shape} {shape.ID}")
        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.ID}")
        assert int(new_shape.ID) > int(shape.ID)  # and new shape has > ID than original
        assert int(new_shape.ID) > max_id
        updated_text = shape.text + " (new copy)"
        new_shape.text = updated_text  # update text of new shape
        assert page.find_shape_by_text(updated_text)
        new_shape_id = new_shape.ID
        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        shape = page.find_shape_by_id(new_shape_id)
        assert shape
        # check that new shape has expected text
        assert shape.text == updated_text


@pytest.mark.parametrize(("filename", "shape_name"),
                         [("test1.vsdx", "Shape to copy"),
                          ("test2.vsdx", "Shape to copy")])
def test_shape_copy_other_page(filename: str, shape_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_shape_copy_other_page.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page =  vis.pages[0]  # type: Page
        page2 = vis.pages[1]  # type: Page
        page3 = vis.pages[2]  # type: Page
        # find and copy shape by name
        shape = page.find_shape_by_text(shape_name)  # type: Shape
        assert shape  # check shape found
        shape_text = shape.text
        print(f"Found shape id:{shape.ID}")

        new_shape = shape.copy(page2)
        assert new_shape  # check copy_shape returns xml
        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.ID}")
        page2_new_shape_id = new_shape.ID

        new_shape = shape.copy(page3)
        assert new_shape  # check copy_shape returns xml
        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.ID}")
        page3_new_shape_id = new_shape.ID

        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        page2 = vis.pages[1]
        shape = page2.find_shape_by_id(page2_new_shape_id)
        assert shape
        assert shape.text == shape_text
        page3 = vis.pages[2]
        shape = page3.find_shape_by_id(page3_new_shape_id)
        assert shape
        assert shape.text == shape_text


# DataProperty

@pytest.mark.parametrize(("filename", "page_index", "shape_name"),
                         [("test1.vsdx", 0, "Shape Text"),
                          ("test6_shape_properties.vsdx", 2, "A")
                          ])
def test_get_shape_data_properties_type_is_dict_of_data_property(filename: str, page_index: int,
                                                                 shape_name: str):
    """ check that Shape.data_properties is a dict of ShapeProperty """
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_name)  # type: Shape

        props = shape.data_properties

        # check overall type
        assert isinstance(props, dict)

        # check each item type
        for prop in props.values():
            assert isinstance(prop, DataProperty)


@pytest.mark.parametrize(("filename", "page_index", "shape_name", "property_dict"),
                         [("test1.vsdx", 0, "Shape Text", {"my_property_label": "property value",
                                                           "my_second_property_label": "another value",
                                                           "Network Name": "Box01"}),
                          ("test6_shape_properties.vsdx", 2, "A", {"master_Prop": "master prop value"}),
                          ("test6_shape_properties.vsdx", 2, "B", {"master_Prop": "master prop value",
                                                                   "shape_prop": "shape property value"}),
                          ("test6_shape_properties.vsdx", 2, "C", {"master_Prop": "override"}),
                          ("test6_shape_properties.vsdx", 2, "D", {"LongProp": 'value not in an "attrib"'}),
                          ])
def test_get_shape_data_properties(filename: str, page_index: int, shape_name: str, property_dict: dict):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_name)  # type: Shape

        props = shape.data_properties
        print(props)
        # check lengths are same
        assert len(props) == len(property_dict)

        # check each key/value is same as expected
        for property_label in props.keys():
            prop = props[property_label]
            print(f"prop: lbl:'{prop.label}' name:'{prop.name}': val:'{prop.value}'")
            assert prop.value == property_dict.get(property_label)


@pytest.mark.parametrize(("filename", "page_index", "shape_name", "property_dict"),
                         [("test1.vsdx", 0, "Shape Text", {"my_property_label": "1",
                                                           "my_second_property_label": "2",
                                                           "Network Name": "3"}),
                          ("test6_shape_properties.vsdx", 2, "A", {"master_Prop": "1"}),
                          ("test6_shape_properties.vsdx", 2, "B", {"master_Prop": "1",
                                                                   "shape_prop": "2"}),
                          ("test6_shape_properties.vsdx", 2, "C", {"master_Prop": "1"}),
                          ("test6_shape_properties.vsdx", 2, "D", {"LongProp": '1"'}),
                          #("test_shape_with_field.vsdx", 0, "Here is field", {"field_label": 'updated field value'}),
                          ])
def test_set_shape_data_properties(filename: str, page_index: int, shape_name: str, property_dict: dict):
    """Check that we can set a prop value, and this change persists through save and load"""
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_set_shape_data_properties.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_name)

        props = shape.data_properties

        # check each key/value is not already set to required value
        for property_label in property_dict.keys():
            prop = props[property_label]
            print(f"prop: lbl:'{prop.label}' name:'{prop.name}': val:'{prop.value}'")
            assert prop.value != property_dict.get(property_label)
            print(f"shape.text={shape.text}")
            print(vis.pretty_print_element(shape.xml))

        # check each key/value is expected after being set
        for property_label in property_dict.keys():
            prop = props[property_label]
            prop.value = property_dict.get(property_label)
            print(f"checking prop after set: lbl:'{prop.label}' name:'{prop.name}': val:'{prop.value}'")
            assert prop.value == property_dict.get(property_label)

        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_name)  # type: Shape
        # check each key/value is not already set to required value
        for property_label in property_dict.keys():
            prop = props[property_label]
            print(f"checking prop after load: lbl:'{prop.label}' name:'{prop.name}': val:'{prop.value}'")
            assert prop.value == property_dict.get(property_label)


@pytest.mark.parametrize(("filename", "page_index", "container_shape_name", "expected_shape_name", "property_label"),
                         [("test6_shape_properties.vsdx", 1, "Container", "Shape A", "label_one"),
                          ])
def test_find_sub_shape_by_data_property_label(filename: str, page_index: int, container_shape_name: str, expected_shape_name: list, property_label: str):
    """Test that we can find a shape and it's text inside a container shape based on the sub shapes labels"""
    with VisioFile(os.path.join(basedir, filename)) as vis:
        container_shape = vis.pages[page_index].find_shape_by_text(container_shape_name)

        shape = container_shape.find_shape_by_property_label(property_label)

        # test that a shape is returned
        assert shape

        # test that the shape returned has the expected names
        assert shape.text.replace('\n', '') == expected_shape_name


@pytest.mark.parametrize(("filename", "page_index", "property_label", "expected_shape_text"),
                          [("test_master_multiple_child_shapes.vsdx", 0, "title",
                            "AWS Step Functions workflow ")
                          ])
def test_find_shape_by_data_property_label(filename: str, page_index: int, property_label: str,
                                           expected_shape_text: str):
    """Test that we can find a shape and it's text inside a page based on the shapes labels"""
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_property_label(property_label)

        # test that a shape is returned
        assert shape

        # test that shape has a DataProperty with expected property label
        assert [prop for prop in shape.data_properties.values() if prop.label == property_label]

        # test that the shape returned has the expected text
        assert shape.text.replace('\n', '') == expected_shape_text


# Connectors

@pytest.mark.parametrize(("filename", "shape_id", "expected_shape_ids"),
                         [
                             ('test4_connectors.vsdx', "1", ["6"]),
                             ('test4_connectors.vsdx', "2", ["6", "7"]),
                          ])
def test_find_connected_shapes(filename: str, shape_id: str, expected_shape_ids: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        actual_connect_shape_ids = list()
        shape = page.find_shape_by_id(shape_id)
        for c_shape in shape.connected_shapes:  # type: Shape
            actual_connect_shape_ids.append(c_shape.ID)
        assert sorted(actual_connect_shape_ids) == sorted(expected_shape_ids)


@pytest.mark.parametrize(("filename", "shape_id", "expected_shape_ids", "expected_from", "expected_to", "expected_from_rels", "expected_to_rels"),
                         [
                             ('test4_connectors.vsdx', "1", ["6"], ["6"], ["1"], ['BeginX'], ['PinX']),
                             ('test4_connectors.vsdx', "2", ["6", "7"], ["6", "7"], ["2", "2"], ['BeginX', 'EndX'], ['PinX', 'PinX']),
                             ('test4_connectors.vsdx', "6", ["1", "2"], ["6", "6"], ["1", "2"], ['BeginX', 'EndX'], ['PinX', 'PinX']),
                             ('test4_connectors.vsdx', "7", ["5", "2"], ["7", "7"], ["5", "2"], ['BeginX', 'EndX'], ['PinX', 'PinX']),
                          ])
def test_find_connected_shape_relationships(filename: str, shape_id: str, expected_shape_ids: list, expected_from: list,
                                            expected_to: list, expected_from_rels: list, expected_to_rels: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page

        shape = page.find_shape_by_id(shape_id)
        shape_ids = [s.ID for s in shape.connected_shapes]
        from_ids = [c.from_id for c in shape.connects]
        to_ids = [c.to_id for c in shape.connects]
        from_rels = [c.from_rel for c in shape.connects]
        to_rels = [c.to_rel for c in shape.connects]

        assert sorted(shape_ids) == sorted(expected_shape_ids)
        assert sorted(from_ids) == sorted(expected_from)
        assert sorted(to_ids) == sorted(expected_to)
        assert sorted(from_rels) == sorted(expected_from_rels)
        assert sorted(to_rels) == sorted(expected_to_rels)


@pytest.mark.parametrize("filename, page_index, shape_text, expected_coords", [
    ("test7_with_connector.vsdx", 0, "Tall Box", [('RelMoveTo', [('X', '0', None), ('Y', '0', None)]), ('RelLineTo', [('X', '1', None), ('Y', '0', None)]), ('RelLineTo', [('X', '1', None), ('Y', '1', None)]), ('RelLineTo', [('X', '0', None), ('Y', '1', None)]), ('RelLineTo', [('X', '0', None), ('Y', '0', None)])]),
    ("test1.vsdx", 0, "Shape Text", [('RelMoveTo', [('X', '0', None), ('Y', '0', None)]), ('RelLineTo', [('X', '1', None), ('Y', '0', None)]), ('RelLineTo', [('X', '1', None), ('Y', '1', None)]), ('RelLineTo', [('X', '0', None), ('Y', '1', None)]), ('RelLineTo', [('X', '0', None), ('Y', '0', None)])]),
    ("test2.vsdx", 2, "Already here", [('Ellipse', [('X', '0.2460629842730571', 'Width*0.5'), ('Y', '0.2519684958956088', 'Height*0.5'), ('A', '0.4921259685461141', 'Width*1'), ('B', '0.2519684958956088', 'Height*0.5'), ('C', '0.2460629842730571', 'Width*0.5'), ('D', '0.5039369917912175', 'Height*1')])]),
    ("test4_connectors.vsdx", 1, "A to B", [('MoveTo', [('X', '0', None), ('Y', '0.09842519685039441', None)]), ('LineTo', [('X', '0.6358267353988465', None), ('Y', '0.09842519685039441', None)])]),
])
def test_get_shape_geometry(filename: str, page_index: str, shape_text: str, expected_coords: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)

        coords = [(r.row_type, [(c.name, c.value, c.func) for c in r.cells.values()]) for r in shape.geometry.rows.values()]
        print(f"coords={coords}")
        print(VisioFile.pretty_print_element(shape.xml))
        if shape.master_shape:
            print(VisioFile.pretty_print_element(shape.master_shape.xml))
        assert coords == expected_coords


@pytest.mark.parametrize("filename_1, page_index_1, shape_text_1, filename_2, page_index_2, shape_text_2, are_equal", [
    ("test1.vsdx", 0, "Shape Text", "test1.vsdx", 0, "Shape Text", True),
    ("test1.vsdx", 0, "Shape Text", "test2.vsdx", 0, "Shape Text", False),
    ])
def test_shape_equality(filename_1, page_index_1, shape_text_1, filename_2, page_index_2, shape_text_2, are_equal):
    with VisioFile(os.path.join(basedir, filename_1)) as vis:
        page = vis.pages[page_index_1]
        shape_1 = page.find_shape_by_text(shape_text_1)

    with VisioFile(os.path.join(basedir, filename_2)) as vis:
        page = vis.pages[page_index_2]
        shape_2 = page.find_shape_by_text(shape_text_2)

    assert (shape_1 == shape_2) == are_equal


@pytest.mark.parametrize("filename, page_index, shape_text, expected_bounds", [
    ("test1.vsdx", 0, "Shape Text", ['0.25', '9.87', '2.42', '11.44']),  # standard shape
    ("test2.vsdx", 0, "Sub-shape 2", ['0.54', '0.17', '1.62', '0.51']),  # sub shape in a group
    ("test2.vsdx", 0, "Scenario:", ['0.73', '8.19', '2.49', '8.97']),  # line
    ])
def test_shape_bounds(filename, page_index, shape_text, expected_bounds):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)
        print([f"{n:.2f}" for n in shape.bounds])
        assert [f"{n:.2f}" for n in shape.bounds] == expected_bounds


@pytest.mark.parametrize("filename, page_index", [
    ("test1.vsdx", 0),
    ("test1.vsdx", 2),
    ("test2.vsdx", 0),
    ])
def test_all_shape_bounds(filename, page_index):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        for s in page.all_shapes:
            # checl all shapes have bounds and relative bounds
            print(s.ID, s.bounds, s.relative_bounds)
            assert s.bounds
            assert s.relative_bounds


@pytest.mark.parametrize("filename, page_index, shape_text, expected_bounds", [
    ("test1.vsdx", 0, "Shape Text", ['0.25', '9.87', '2.42', '11.44']),  # standard shape
    ("test2.vsdx", 0, "Sub-shape 2", ['0.79', '10.04', '1.87', '10.37']),  # sub shape in a group
    ("test2.vsdx", 0, "Scenario:", ['0.73', '8.19', '2.49', '8.97']),  # line
    ])
def test_shape_relative_bounds(filename, page_index, shape_text, expected_bounds):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)
        print([f"{n:.2f}" for n in shape.relative_bounds])
        assert [f"{n:.2f}" for n in shape.relative_bounds] == expected_bounds


@pytest.mark.parametrize("filename, page_index, shape_text, arrow", [
    ("test2.vsdx", 0, "Scenario:", True),  # add end arrow
    ("test2.vsdx", 0, "Scenario:", False),  # no end arrow
    ])
def test_shape_end_arrow(filename, page_index, shape_text, arrow):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_shape_end_arrow_{arrow}.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)
        shape.end_arrow = arrow
        print(shape.end_arrow)
        print(vsdx.pretty_print_element(shape.xml))
        assert shape.end_arrow == '13' if arrow else '0'
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)
        print(shape.end_arrow)
        print(vsdx.pretty_print_element(shape.xml))
        assert shape.end_arrow == '13' if arrow else '0'


@pytest.mark.parametrize("filename, page_index, shape_text, expected_universal_name", [
    ("test3_house.vsdx", 0, "context filter", "House"),
    ("test4_connectors.vsdx", 1, "Shape A", None),  # no master and no name
    ("test4_connectors.vsdx", 1, "A to B", "Dynamic connector"),  # master with Layer and UnivName Cell
    ("test4_connectors.vsdx", 2, "Switch", "Switch"),  # master with no Layer
    ("test4_connectors.vsdx", 2, "Router", "Router"),  # master with no Layer
    ])
def test_shape_universal_name(filename, page_index, shape_text, expected_universal_name):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        shape = page.find_shape_by_text(shape_text)
        assert shape.universal_name == expected_universal_name


@pytest.mark.parametrize("filename, expected_master_shape_name",
                         [("test_master_multiple_child_shapes.vsdx",  # master with multiple child shapes
                           "AWS Step Functions workflow ")])  # expected master shape name
def test_get_shape_master_page(filename: str, expected_master_shape_name: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        child_shape = vis.get_page(0).child_shapes[0]
        assert child_shape.master_page.name == expected_master_shape_name


@pytest.mark.parametrize("filename, shape_id, expected_angle",
                         [
                             ("test11_rotate.vsdx", "1", 0.52),
                             ("test11_rotate.vsdx", "2", -1.39),
                             ("test11_rotate.vsdx", "5", 2.53),
                             ("test11_rotate.vsdx", "6", -0.26),
                         ])
def test_get_shape_angle(filename: str, shape_id: str, expected_angle: float):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]
        # check angle is close enough to expected
        assert abs(page.find_shape_by_id(shape_id).angle - expected_angle) < 0.01


@pytest.mark.parametrize(("filename", "regex", "expected_shape_ids"),
                         [
                             ('test1.vsdx', r'\s(\S{2})\s', ['2', '5', '6']),
                             
                          ])
def test_find_shapes_by_regex(filename: str, regex: str, expected_shape_ids: list):
    """ Test function Shape.find_shapes_by_regex(regex: str) 
        test1.vsdx contains Shapes with text fields
            [(shp.ID,shp.text) for shp in shapes.all_shapes]:
            ('1', 'Shape Text\n')
            ('2', 'Shape to remove\n')
            ('5', 'Shape to copy\n')
            ('6', 'Shape for context filter: The scenario is {{scenario}} and this file was created on {{date}}\n')
        
    the regex matches the sequence '(whitespace)(two non-whitespace)(whitespace)' as for example with the strings ' to ' in ID=2,5 and also ' is ' and ' on ' in ID=6
    so we expect IDs=[2,5,6]
    """
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shapes = vis.pages[0].shapes[0]  # type: Page
        assert len(shapes.find_shapes_by_regex('')) == len(shapes.all_shapes)
        fil_shapes = shapes.find_shapes_by_regex(regex)
        assert [shp.ID for shp in fil_shapes] == expected_shape_ids


@pytest.mark.parametrize(("filename", "page_index", "shape_text", "color_param", "expected_colour"),
                         [
                             ('test12_colors.vsdx', 0, "Line Color", "line", "#ff0000"),
                             ('test12_colors.vsdx', 0, "Text Color", "text", "#ff0000"),
                             ('test12_colors.vsdx', 0, "Fill Color", "fill", "#ff0000"),
                             ('test12_colors.vsdx', 1, "Line Color", "line", "#00ff00"),
                             ('test12_colors.vsdx', 1, "Text Color", "text", "#00ff00"),
                             ('test12_colors.vsdx', 1, "Fill Color", "fill", "#00ff00"),

                         ])
def test_get_shape_line_color(filename: str, page_index: int, shape_text: str, color_param: str, expected_colour: str):
    """Test that we can get a shapes line, text, or fill color"""
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_text)
        print(f"LineColor={shape.line_color} FillColor={shape.fill_color} TextColor={shape.text_color}")
        if color_param == "line":
            assert shape.line_color == expected_colour
        elif color_param == "text":
            assert shape.text_color == expected_colour
        elif color_param == "fill":
            assert shape.fill_color == expected_colour


@pytest.mark.parametrize(("filename", "page_index", "shape_text", "color_param", "expected_colour"),
                         [
                             ('test12_colors.vsdx', 0, "Line Color", "line", "#00ff00"),
                             ('test12_colors.vsdx', 0, "Text Color", "text", "#00ff00"),
                             ('test12_colors.vsdx', 0, "Fill Color", "fill", "#00ff00"),
                             ('test12_colors.vsdx', 1, "Line Color", "line", "#0000ff"),
                             ('test12_colors.vsdx', 1, "Text Color", "text", "#0000ff"),
                             ('test12_colors.vsdx', 1, "Fill Color", "fill", "#0000ff"),

                         ])
def test_set_shape_line_color(filename: str, page_index: int, shape_text: str, color_param: str, expected_colour: str):
    """Test that we can set a shapes line, text, or fill color"""
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_{page_index}_set_shape_{color_param}_color.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_text)
        print(f"LineColor={shape.line_color} FillColor={shape.fill_color} TextColor={shape.text_color}")
        if color_param == "line":
           shape.line_color = expected_colour
        elif color_param == "text":
            shape.text_color = expected_colour
        elif color_param == "fill":
            shape.fill_color = expected_colour
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        shape = vis.pages[page_index].find_shape_by_text(shape_text)
        print(f"LineColor={shape.line_color} FillColor={shape.fill_color} TextColor={shape.text_color}")
        if color_param == "line":
            assert shape.line_color == expected_colour
        elif color_param == "text":
            assert shape.text_color == expected_colour
        elif color_param == "fill":
            assert shape.fill_color == expected_colour



import wsgiref.headers

import pytest
from datetime import datetime
import os

from typing import List

import vsdx
from vsdx import Connect
from vsdx import Page
from vsdx import Shape
from vsdx import VisioFile
from vsdx import Cell

# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))


@pytest.mark.parametrize("filename, count", [("test1.vsdx", 1), ("test2.vsdx", 1)])
def test_get_page_shapes(filename: str, count: int):
    # there should always be one single Shapes object in the page that contains Shape objects
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        assert len(page._shapes) == count  # _shapes() is internal function to get container object


@pytest.mark.parametrize("filename, child_count",
                         [("test1.vsdx", 4),
                          ("test2.vsdx", 6),
                          ("test10_nested_shapes.vsdx", 2)
                          ])
def test_get_page_child_shapes(filename: str, child_count: int):
    # Check that page has expected number of top level shapes
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        # check that page has expected number of child shapes
        assert len(page.child_shapes) == child_count
        # check that same number of shapes can be found in _shapes and page
        assert len(page._shapes[0].child_shapes) == len(page.child_shapes)


@pytest.mark.parametrize("filename, all_count",
                         [("test1.vsdx", 4),
                          ("test2.vsdx", 14),
                          ("test10_nested_shapes.vsdx", 8),
                          ])
def test_get_page_all_shapes(filename: str, all_count: int):
    # Check that page has expected number of total shapes (all_shapes searches recursively)
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        # check that page has expected number of all_shapes
        print(page.all_shapes)
        assert len(page.all_shapes) == all_count
        # check that same number of shapes can be found in _shapes and page
        assert len(page._shapes[0].all_shapes) == len(page.all_shapes)


@pytest.mark.parametrize("filename, page_index, height_width",
                         [("test1.vsdx", 0, (8.26771653543307, 11.69291338582677)),
                          ("test2.vsdx", 0, (8.26771653543307, 11.69291338582677)),
                          ("test1.vsdx", 1, (8.26771653543307, 11.69291338582677)),
                          ("test2.vsdx", 1, (8.26771653543307, 11.69291338582677)),])
def test_get_page_size(filename: str, page_index: int, height_width: tuple):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        print(VisioFile.pretty_print_element(page._pagesheet_xml))
        print(f"\n w x h={page.width} x {page.height}")
        assert (page.width, page.height) == height_width


@pytest.mark.parametrize("filename, page_index, page_scale",
                         [("test1.vsdx", 0, 0.5),
                          ("test2.vsdx", 0, 0.5),
                          ("test1.vsdx", 1, 0.5),
                          ("test2.vsdx", 1, 0.5),])
def test_set_page_size(filename: str, page_index: int, page_scale: float):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_set_page_size_{page_index}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        #print(VisioFile.pretty_print_element(page._pagesheet_xml))
        print(f"\n w x h={page.width} x {page.height}")
        page_width = page.width * page_scale
        page_height = page.height * page_scale
        page.width = page_width
        page.height = page_height
        print(f"\n w x h={page.width} x {page.height}")
        vis.save_vsdx(out_file)

        with VisioFile(out_file) as vis:
            page = vis.pages[page_index]
            print(f"\n w x h={page.width} x {page.height}")
            assert page.width == page_width
            assert page.height == page_height


@pytest.mark.parametrize("filename, page_index, expected_shape_bounds",
                         [("test1.vsdx", 0, (0, 0, 1, 1)),
                          ("test2.vsdx", 0, (0, 0, 1, 1)),
                          ("test1.vsdx", 2, (0, 0, 1, 1)),
                          ("test2.vsdx", 2, (0, 0, 1, 1)),])
def test_get_page_bounds(filename: str, page_index: int, expected_shape_bounds: tuple):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_get_page_bounds_{page_index}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]

        box = vsdx.media.Media().rectangle

        for s in page.all_shapes:
            #bx = s.begin_x or (s.x-s.loc_x)
            #by = s.begin_y or (s.y-s.loc_y)
            #ex = s.end_x or (bx + s.width)
            #ey = s.end_y or (by + s.height)
            bx, by, ex, ey = s.bounds
            #print(f"{s.ID} bx:{bx} by:{by} ex:{ex} ey:{ey} x:{s.x} y:{s.y} lx:{s.loc_x} ly:{s.loc_y} w:{s.width} h:{s.height} {s.text} {s.geometry.rows if s.geometry else None}")
            cbox = box.copy(page=page)
            #print(vsdx.pretty_print_element(list(cbox.cells.values())[0].xml))
            print(s.text, s.bounds, s.parent.shape_type)
            cbox.line_color = '#ff2222'
            cbox.x = bx
            cbox.loc_x = 0
            cbox.width = ex-bx
            cbox.y = by
            cbox.loc_y = 0
            cbox.height = ey-by
            cbox.text = f"{bx:.2g},{by:.2g}-{ex:.2g},{ey:.2g}"
            #print(vsdx.pretty_print_element(cbox.xml))
        vis.save_vsdx(out_file)


@pytest.mark.parametrize("filename, page_index, page_name",
                         [("test1.vsdx", 0, "Page-1"),
                          ("test2.vsdx", 0, "Page-1"),
                          ("test1.vsdx", 1, "Page-2"),
                          ("test2.vsdx", 1, "Page-2"),])
def test_get_page_name(filename: str, page_index: int, page_name: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        print(VisioFile.pretty_print_element(page._pagesheet_xml))
        assert page.name == page_name


@pytest.mark.parametrize("filename, page_index, page_name",
                         [("test1.vsdx", 0, "Page1"),
                          ("test2.vsdx", 0, "Page1"),
                          ("test1.vsdx", 1, "Page2"),
                          ("test2.vsdx", 1, "Page2"),])
def test_set_page_name(filename: str, page_index: int, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_set_page_name_{page_name}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        print(VisioFile.pretty_print_element(page._pagesheet_xml))
        page.name = page_name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.pages[page_index]
        assert page.name == page_name


@pytest.mark.parametrize("filename, count", [("test1.vsdx", 4), ("test2.vsdx", 6)])
def test_get_page_sub_shapes(filename: str, count: int):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shapes = page.child_shapes
        print(f"shape count={len(shapes)}")
        assert len(shapes) == count


@pytest.mark.parametrize("filename, shape_id", [("test1.vsdx", "6"), ("test2.vsdx", "6")])
def test_get_shape_with_text(filename: str, shape_id: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_text('{{date}}')  # type: Shape
        assert shape.ID == shape_id


@pytest.mark.parametrize("filename", ["test1.vsdx", "test2.vsdx", "test3_house.vsdx"])
def test_apply_context(filename: str):
    date_str = str(datetime.today().date())
    context = {'scenario': 'test',
               'date': date_str}
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_VISfilter_applied.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        original_shape = page.find_shape_by_text('{{date}}')  # type: Shape
        assert original_shape.ID
        page.apply_text_context(context)
        vis.save_vsdx(out_file)

    # open and find date_str
    with VisioFile(out_file) as vis:
        page = vis.get_page(0)  # type: Page
        updated_shape = page.find_shape_by_text(date_str)  # type: Shape
        assert updated_shape.ID == original_shape.ID


@pytest.mark.parametrize("filename", ["test1.vsdx", "test2.vsdx", "test3_house.vsdx"])
def test_find_replace(filename: str):
    old = 'Shape'
    new = 'Figure'
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_VISfind_replace_applied.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        original_shapes = page.find_shapes_by_text(old)  # type: List[Shape]
        shape_ids = [s.ID for s in original_shapes]
        page.find_replace(old, new)
        vis.save_vsdx(out_file)

    # open and find date_str
    with VisioFile(out_file) as vis:
        page = vis.get_page(0)  # type: Page
        # test that each shape if has 'new' str in text
        for shape_id in shape_ids:
            shape = page.find_shape_by_id(shape_id)
            assert new in shape.text


# DataProperty

@pytest.mark.parametrize(("filename", "page_index", "expected_shape_name", "property_label"),
                         [("test1.vsdx", 0, "Shape Text", "my_property_label"),
                          ("test6_shape_properties.vsdx", 0, "Shape One", "my_property_label"),
                          ("test6_shape_properties.vsdx", 0, "Shape One", "my_second_property_label"),
                          ("test6_shape_properties.vsdx", 0, "Shape Three", "my_third_property_label"),
                          ("test6_shape_properties.vsdx", 2, "A", "master_Prop"),
                          ])
def test_find_shape_by_data_property_label(filename: str, page_index: int, expected_shape_name: str, property_label: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_property_label(property_label)  # type: Shape

        # test that a shape is returned
        assert shape

        # test that returned shape has a DataProperty with property_name
        assert shape.data_properties[property_label]

        # test iot's the expected shape
        assert shape.text.replace('\n', '') == expected_shape_name


@pytest.mark.parametrize(("filename", "page_index", "expected_shape_names", "property_label"),
                         [("test1.vsdx", 0, ["Shape Text"], "my_property_label"),
                          ("test6_shape_properties.vsdx", 0, ["Shape One", "Shape Two"], "my_property_label"),
                          ("test6_shape_properties.vsdx", 0, ["Shape One", "Shape Three"], "my_second_property_label"),
                          ("test6_shape_properties.vsdx", 1, ["Shape A", "Shape C"], "label_one"),
                          ("test6_shape_properties.vsdx", 2, ["A", "B", "C"], "master_Prop"),
                          ])
def test_find_shapes_by_data_property_label(filename: str, page_index: int, expected_shape_names: list, property_label: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shapes = vis.pages[page_index].find_shapes_by_property_label(property_label)

        # test that a shape is returned
        assert len(shapes) == len(expected_shape_names)

        # get list of shape names (text) without carriage returns
        shape_names = sorted([s.text.replace('\n', '') for s in shapes])
        print(f"shape names={shape_names}")

        # test that the shapes returned have the expected names
        assert shape_names == expected_shape_names

        # test that each returned shape has a DataProperty with property_name
        for s in shapes:
            assert s.data_properties[property_label]


@pytest.mark.parametrize(("filename", "page_index", "expected_shape_names", "property_label", "property_value"),
                         [("test1.vsdx", 0, ["Shape Text"], "my_property_label", "property value"),
                          ("test6_shape_properties.vsdx", 0, ["Shape One"], "my_property_label", "property value"),
                          ("test6_shape_properties.vsdx", 0, ["Shape Three"], "my_second_property_label", "a different value"),
                          ("test6_shape_properties.vsdx", 1, ["Shape A", "Shape C"], "label_one", "0"),
                          ("test6_shape_properties.vsdx", 2, ["A", "B"], "master_Prop", "master prop value"),
                          ("test6_shape_properties.vsdx", 2, ["C"], "master_Prop", "override"),
                          ])
def test_find_shapes_by_data_property_label_value(filename: str, page_index: int, expected_shape_names: list, property_label: str, property_value: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shapes = vis.pages[page_index].find_shapes_by_property_label_value(property_label, property_value)

        # test that a shape is returned
        assert len(shapes) == len(expected_shape_names)

        # get list of shape names (text) without carriage returns
        shape_names = sorted([s.text.replace('\n', '') for s in shapes])
        print(f"shape names={shape_names}")

        # test that the shapes returned have the expected names
        assert shape_names == expected_shape_names

        # test that each returned shape has a DataProperty with property_name
        for s in shapes:
            assert s.data_properties[property_label]
            print(f"{property_label}='{s.data_properties[property_label].value}'")


@pytest.mark.parametrize(("filename", "page_index", "expected_shape_name", "property_label", "property_value"),
                         [("test1.vsdx", 0, "Shape Text", "my_property_label", "property value"),
                          ("test6_shape_properties.vsdx", 0, "Shape One", "my_property_label", "property value"),
                          ("test6_shape_properties.vsdx", 0, "Shape Three", "my_second_property_label", "a different value"),
                          ("test6_shape_properties.vsdx", 1, "Shape A", "label_one", "0"),
                          ("test6_shape_properties.vsdx", 2, "A", "master_Prop", "master prop value"),
                          ("test6_shape_properties.vsdx", 2, "C", "master_Prop", "override"),
                          ])
def test_find_shape_by_data_property_label_value(filename: str, page_index: int, expected_shape_name: str, property_label: str, property_value: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        shape = vis.pages[page_index].find_shape_by_property_label_value(property_label, property_value)

        # test that the shapes returned have the expected names
        assert shape.text.replace('\n', '') == expected_shape_name


@pytest.mark.parametrize(("filename", "page_index", "regex", "expected_shape_ids"),
                         [
                             ('test1.vsdx', 0, r'\s(\S{2})\s', ['2', '5', '6']),

                         ])
def test_find_shapes_by_regex(filename: str, page_index, regex: str, expected_shape_ids: list):
    """ Test function Shape.find_shapes_by_regex(regex: str)
        test1.vsdx contains Shapes with text fields
            [(shp.ID,shp.text) for shp in shapes.all_shapes]:
            ('1', 'Shape Text\n')
            ('2', 'Shape to remove\n')
            ('5', 'Shape to copy\n')
            ('6', 'Shape for context filter: The scenario is {{scenario}} and this file was created on {{date}}\n')

    the regex matches the sequence '(whitespace)(two non-whitespace)(whitespace)' as for example with the strings ' to ' in ID=2,5 and also ' is ' and ' on ' in ID=6
    so we expect IDs=[2,5,6]
    """
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]  # type: Page
        assert len(page.find_shapes_by_regex('')) == len(page.all_shapes)
        fil_shapes = page.find_shapes_by_regex(regex)
        assert [shp.ID for shp in fil_shapes] == expected_shape_ids


# Connectors

@pytest.mark.parametrize(("filename", "expected_connects"),
                         [('test4_connectors.vsdx', ["from 7 to 5", "from 7 to 2", "from 6 to 2", "from 6 to 1"]),
                          ])
def test_find_page_connects(filename: str, expected_connects: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        actual_connects = list()
        for c in page.connects:  # type: Connect
            actual_connects.append(f"from {c.from_id} to {c.to_id}")
        assert sorted(actual_connects) == sorted(expected_connects)


@pytest.mark.parametrize(("filename", "shape_a_id", "shape_b_id", "expected_connector_ids"),
                         [
                             ('test4_connectors.vsdx', "1", "2", ["6"]),
                             ('test4_connectors.vsdx', "1", "5", []),  # expect no connections
                          ])
def test_find_connectors_between_ids(filename: str, shape_a_id: str, shape_b_id: str,  expected_connector_ids: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        connectors = page.get_connectors_between(shape_a_id=shape_a_id, shape_b_id=shape_b_id)
        actual_connector_ids = sorted([c.ID for c in connectors])
        assert sorted(expected_connector_ids) == list(actual_connector_ids)


@pytest.mark.parametrize(("filename", "shape_a_text", "shape_b_text", "expected_connector_ids"),
                         [
                             ('test4_connectors.vsdx', "Shape A", "Shape B", ["6"]),
                             ('test4_connectors.vsdx', "Shape A", "Shape C", []),  # expect no connections
                          ])
def test_find_connectors_between_shapes(filename: str, shape_a_text: str, shape_b_text: str,  expected_connector_ids: list):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        connectors = page.get_connectors_between(shape_a_text=shape_a_text, shape_b_text=shape_b_text)
        actual_connector_ids = sorted([c.ID for c in connectors])
        assert sorted(expected_connector_ids) == list(actual_connector_ids)


@pytest.mark.parametrize(("filename", "page_index", "shape_a_text", "shape_b_text"),
                         [
                             ('test8_simple_connector.vsdx', 0, "Shape A", "Shape B"),
                             ('test7_with_connector.vsdx', 0, "Shape Text", "Shape to remove"),
                             ('test1.vsdx', 0, "Shape to remove", "Shape Text"),
                             ('test1.vsdx', 0, "Shape Text", "Shape to copy"),
                             ('test1.vsdx', 0, "Shape to copy", "Shape to remove"),
                             ('test4_connectors.vsdx', 0, "Shape A", "Shape C"),
                             ('test4_connectors.vsdx', 0, "Shape C", "Shape B"),
                          ])
def test_add_connect_between_shapes(filename: str, page_index: int, shape_a_text: str, shape_b_text: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_add_connect_between_'+shape_a_text+'_'+shape_b_text+'.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"filename:{out_file}")
        page = vis.pages[page_index]  # type: Page
        from_shape = page.find_shape_by_text(shape_a_text)
        to_shape = page.find_shape_by_text(shape_b_text)
        c = Connect.create(page=page, from_shape=from_shape, to_shape=to_shape)
        c.end_arrow = True
        new_connector_id = c.ID

        c.text = "NEW"
        vis.save_vsdx(out_file)

        # re-open saved file and check it is changed as expected
        with VisioFile(out_file) as vis:
            page = vis.pages[page_index]
            connector_ids = [c.connector_shape_id for c in page.connects]
            # new shape is referenced as connector_shape for new connector relationship
            assert new_connector_id in connector_ids
            # new shape exists in page
            c = page.find_shape_by_id(new_connector_id)
            assert page.find_shape_by_id(new_connector_id)


@pytest.mark.parametrize(("filename"),
                         [
                             ('test8_simple_connector.vsdx'),
                          ])
def test_add_multiple_connectors(filename: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        src_page = vis.pages[0]
        block_shape = src_page.child_shapes[0]
        new_page = vis.add_page('new page')
        # 1
        # adding new_shape1, new_shape2 and connecting them
        new_shape1 = block_shape.copy(new_page)
        new_shape1.text = 'new shape 1'
        new_shape2 = block_shape.copy(new_page)
        new_shape2.text = 'new shape 2'
        Connect.create(page=new_page, from_shape=new_shape1, to_shape=new_shape2)
        out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_new_test_1.vsdx')
        vis.save_vsdx(out_file)
        # 2
        # adding new_shape3 and connecting it to new_shape2
        new_shape3 = block_shape.copy(new_page)
        new_shape3.text = 'new shape 3'
        Connect.create(page=new_page, from_shape=new_shape2, to_shape=new_shape3)  # ERROR
        out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_new_test_2.vsdx')
        vis.save_vsdx(out_file)

@pytest.mark.parametrize(("filename", "page_index", "shape_a_label", "shape_a_value", "shape_b_label", "shape_b_value"),
                         [
                             ('test1.vsdx', 0, "Network Name", "Box01", "Network Name", "Box02"),
                             ('test3_house.vsdx', 0, "Network Name", "House01", "Network Name", "Box01"),
                             ('test4_connectors.vsdx', 2, "Network Name", "Box01", "Network Name", "Box02"),
                             ('test4_connectors.vsdx', 2, "Network Name", "Box01", "Network Name", "Router01"),
                             ('test4_connectors.vsdx', 2, "Network Name", "Switch01", "Network Name", "Router01"),
                          ])
def test_add_connect_between_shapes_by_property(filename: str, page_index: int, shape_a_label: str, shape_a_value,
                                                shape_b_label: str, shape_b_value: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_add_connect_between_labels_'+shape_a_value+'_'+shape_b_value+'.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"filename:{out_file}")
        page = vis.pages[page_index]  # type: Page
        from_shape = page.find_shape_by_property_label_value(shape_a_label, shape_a_value)
        to_shape = page.find_shape_by_property_label_value(shape_b_label, shape_b_value)

        c = Connect.create(page=page, from_shape=from_shape, to_shape=to_shape)
        c.end_arrow = True
        new_connector_id = c.ID

        c.text = "NEW"
        vis.save_vsdx(out_file)

        # re-open saved file and check it is changed as expected
        with VisioFile(out_file) as vis:
            page = vis.pages[page_index]
            connector_ids = [c.connector_shape_id for c in page.connects]
            # new shape is referenced as connector_shape for new connector relationship
            assert new_connector_id in connector_ids
            # new shape exists in page
            c = page.find_shape_by_id(new_connector_id)
            assert page.find_shape_by_id(new_connector_id)


def fl(v: float):
    if type(v) is float:
        return f"{v:.2g}"


def add_shape_info(s: Shape):
    s.text = s.text + f" x,y={fl(s.x)}, {fl(s.y)}"
    if s.begin_x is not None:
        s.text = s.text + f" bx,by={fl(s.begin_x)}, {fl(s.begin_y)}"
        s.text = s.text + f" ex,ey={fl(s.end_x)}, {fl(s.end_y)}"
    s.text = s.text + f" w,h={fl(s.width)}, {fl(s.height)}"
    cx, cy = s.center_x_y
    s.text = s.text + f" cx,cy={fl(cx)}, {fl(cy)}"
    s.text = s.text + f" locx,locy={fl(s.loc_x)}, {fl(s.loc_y)}"


@pytest.mark.parametrize(("filename", "shape_text", "lx", "ly"),
                         [
                             ('test9_rect_and_line.vsdx', "Rect A", 2.0, 8.0),
                             ('test9_rect_and_line.vsdx', "Line A", 2.0, 8.0),
                             ('test9_rect_and_line.vsdx', "Conn A", 2.0, 8.0),
                             ('test9_rect_and_line.vsdx', "Rect A", 2.0, 5.0),
                             ('test9_rect_and_line.vsdx', "Line A", 2.0, 5.0),
                             ('test9_rect_and_line.vsdx', "Conn A", 2.0, 5.0),
                          ])
def test_copy_and_move_shape(filename: str, shape_text: str, lx: float, ly: float):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_and_move_shape_{shape_text}_{lx}_{ly}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"filename:{out_file}")
        page = vis.pages[0]  # type: Page
        shape = page.find_shape_by_text(shape_text)
        print(vsdx.pretty_print_element(shape.xml))
        # should work for shapes
        cp1 = shape.copy()
        add_shape_info(shape)
        cp1.x, cp1.y = lx, ly
        cp1.text = cp1.text + f"lx,ly={fl(lx)}, {fl(ly)}"
        if shape.begin_x is not None:
            # should work for lines
            cp1.begin_x, cp1.begin_y = lx, ly
            cp1.end_x, cp1.end_y = lx + cp1.width, ly + cp1.height
            cp1.x, cp1.y = cp1.center_x_y
        add_shape_info(cp1)

        vis.save_vsdx(out_file)


@pytest.mark.parametrize(("filename", "shape_text", "start", "finish"),
                         [
                             ('test9_rect_and_line.vsdx', "Line A", (2.0, 7.0), (3.0, 8.0)),
                             ('test9_rect_and_line.vsdx', "Conn A", (2.0, 7.0), (3.0, 8.0)),
                          ])
def test_copy_and_move_line(filename: str, shape_text: str, start: tuple, finish: tuple):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_and_move_line_{shape_text}_{start}_{finish}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"filename:{out_file}")
        page = vis.pages[0]  # type: Page
        shape = page.find_shape_by_text(shape_text)
        print(vsdx.pretty_print_element(shape.xml))
        # should work for shapes
        cp1 = shape.copy()
        add_shape_info(shape)
        cp1.x, cp1.y = start
        cp1.text = cp1.text + f"lx,ly={fl(cp1.x)}, {fl(cp1.y)}"
        if cp1.begin_x is not None:
            is_connector = cp1.shape_name == 'Dynamic connector'
            # should work for lines
            cp1.begin_x, cp1.begin_y = start
            cp1.end_x, cp1.end_y = finish
            cp1.width = cp1.end_x - cp1.begin_x
            if is_connector:
                cp1.height = cp1.end_y - cp1.begin_y  # connector has a height
            else:
                cp1.height = 0.0  # line height is always zero
            cp1.x, cp1.y = start # cp1.center_x_y
            cp1.geometry.set_move_to(0.0, 0.0)
            cp1.geometry.set_line_to(cp1.width, cp1.height)
            txt_pin_x = cp1.cells.get('TxtPinX')
            txt_pin_y = cp1.cells.get('TxtPinY')
            if txt_pin_x and txt_pin_y:
                if is_connector:
                    txt_pin_x.value, txt_pin_y.value = cp1.width/2, cp1.height/2
                else:
                    txt_pin_x.value, txt_pin_y.value = cp1.center_x_y
                cp1.set_cell_value(name='Control/TextPosition/X', value=txt_pin_x.value)
                cp1.set_cell_value(name='Control/TextPosition/Y', value=txt_pin_y.value)
                cp1.set_cell_value(name='Control/TextPosition/XDyn', value=txt_pin_x.value)
                cp1.set_cell_value(name='Control/TextPosition/YDyn', value=txt_pin_y.value)
                #print(cp1.cells.keys())
            cells = list(cp1.cells.values())+cp1.geometry.cells
            for r in cp1.geometry.rows.values():
                cells.extend(r.cells.values())
            #print(cells)
            for c in cells:  # type: Cell
                v = None
                formula = c.formula
                if formula:
                    if formula == 'Inh' and cp1.master_shape:
                        print(f'Inh: {c.name} {cp1.master_shape.cells.get(c.name)}')
                        master_c = cp1.master_shape.cells.get(c.name)
                        formula = master_c.formula if master_c else formula
                    v = vsdx.calc_value(cp1, formula)
                    if v is not None:
                        c.value = v
                #print(f"c={c.name} f={c.formula} v={v}")

            print(vsdx.pretty_print_element(cp1.xml))
        add_shape_info(cp1)

        vis.save_vsdx(out_file)


@pytest.mark.parametrize("filename, page_index, expected_ids", [
    ("test1.vsdx", 0, ['1', '2', '5', '6']),
    ("test1.vsdx", 1, []),  # empty page
    ("test1.vsdx", 2, ['1']),
    ("test2.vsdx", 0, ['6', '9', '1', '7', '8', '11', '2', '10', '14', '5', '12', '13', '16', '17']),
    ("test2.vsdx", 1, []),  # empty page
    ("test2.vsdx", 2, ['1', '2', '3', '4']),
    ])
def test_page_all_shapes(filename, page_index, expected_ids):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        # all_shapes() gets all shapes on a page, recursively
        shape_ids = [s.ID for s in page.all_shapes]
        print(shape_ids)
        assert shape_ids == expected_ids


@pytest.mark.parametrize("filename, page_index, expected_page_id", [
    ("test1.vsdx", 0, "0"), ("test1.vsdx", 1, "4"), ("test1.vsdx", 2, "7"),
    ("test2.vsdx", 0, "0"), ("test2.vsdx", 1, "4"), ("test2.vsdx", 2, "5"),
    ("test4_connectors.vsdx", 0, "0"), ("test4_connectors.vsdx", 1, "4"), ("test4_connectors.vsdx", 2, "7"),
    ])
def test_page_id(filename, page_index, expected_page_id):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index]
        assert page.page_id == expected_page_id


@pytest.mark.parametrize("filename, master_index, expected_page_id", [
    ("test3_house.vsdx", 0, "2"),
    ("test5_master.vsdx", 0, "1"),
    ("test4_connectors.vsdx", 0, "2"), ("test4_connectors.vsdx", 1, "6"), ("test4_connectors.vsdx", 2, "7"),
    ])
def test_master_page_id(filename, master_index, expected_page_id):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.master_pages[master_index]
        assert page.page_id == expected_page_id


@pytest.mark.parametrize("filename", [
    ("test1.vsdx"),
    ("test2.vsdx"),
    ("test4_connectors.vsdx"),
    ])
def test_page_has_page_sheet_xml(filename):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        for page in vis.pages:
            assert page._pagesheet_xml is not None


@pytest.mark.parametrize("filename", [
    ("test3_house.vsdx"),
    ("test5_master.vsdx"),
    ("test4_connectors.vsdx"),
    ])
def test_master_page_has_sheet_xml(filename):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        for page in vis.master_pages:
            print(page.page_id, page.page_name)
            assert page._pagesheet_xml is not None



"""Pytest Tests for VisioFile class"""
import os
import pytest

from vsdx import ext_prop_namespace
from vsdx import namespace
from vsdx import vt_namespace

from vsdx import Media
from vsdx import Page
from vsdx import PagePosition
from vsdx import Shape
from vsdx import VisioFile

# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))

# file structure


def test_invalid_file_type():
    """Test that opening an invalid file name results in a TypeError"""
    filename = __file__
    print(f"Opening invalid but existing {filename}")
    try:
        with VisioFile(filename):
            assert False  # don't expect to get here
    except TypeError as error:
        print(type(error), error)  # expect to get here


def test_file_closure():
    filename = os.path.join(basedir, 'test1.vsdx')
    directory = f"./{filename.rsplit('.', 1)[0]}"
    with VisioFile(filename):
        # confirm directory exists
        assert os.path.exists(directory)
    # confirm directory is gone
    assert not os.path.exists(directory)


@pytest.mark.parametrize("filename", [
    "test1.vsdx",
    "diagram_with_macro.vsdm",
])
def test_open_rel_path(filename: str):
    # test opening a file in tests directory with relative path
    filename = os.path.join(basedir, filename)
    assert os.path.exists(filename)
    with VisioFile(filename) as vis:
        for page in vis.pages:
            print(page.name)


def test_open_abs_path():
    # test opening media file (not in tests directory)with absolute path
    media_file_path = Media()._media_vsdx.filename
    filename = os.path.abspath(media_file_path)

    assert os.path.exists(filename)
    with VisioFile(filename) as vis:
        for page in vis.pages:
            print(page.name)


def test_open_abs_path_save_rel_path():
    # test opening media file (not in tests directory)with absolute path
    media_file_path = Media()._media_vsdx.filename
    filename = os.path.abspath(media_file_path)

    assert os.path.exists(filename)
    output_file = os.path.join(basedir, 'out', 'abs_to_rel_out.vsdx')
    with VisioFile(filename) as vis:
        vis.save_vsdx(output_file)


def test_open_abs_path_save_abs_path():
    # test opening media file (not in tests directory)with absolute path
    media_file_path = Media()._media_vsdx.filename
    filename = os.path.abspath(media_file_path)

    assert os.path.exists(filename)
    output_file = os.path.abspath(os.path.join(basedir, 'out', 'abs_to_abs_out.vsdx'))
    print('output_file', output_file)
    with VisioFile(filename) as vis:
        vis.save_vsdx(output_file)

# Helpers


@pytest.mark.parametrize(("filename", "shape_elements"), [
    ("test1.vsdx", 4),
    ("test2.vsdx", 14),
    ("test3_house.vsdx", 10)
])
def test_xml_findall_shapes(filename: str, shape_elements: int):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        # find all Shape elements
        xml = page.xml.getroot()
        xpath = f".//{namespace}Shape"
        elements = xml.findall(xpath)
        print(f"{xpath} returns {len(elements)} elements vs {shape_elements}")
        assert len(elements) == shape_elements


@pytest.mark.parametrize(("filename", "group_shape_elements"), [
    ("test1.vsdx", 0),
    ("test2.vsdx", 3),
    ("test3_house.vsdx", 2)
])
def test_xml_findall_group_shapes(filename: str, group_shape_elements: int):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        # find all Shape elements where attribute Type='Group'
        xml = page.xml.getroot()
        xpath = f".//{namespace}Shape[@Type='Group']"
        elements = xml.findall(xpath)
        print(f"{xpath} returns {len(elements)} elements vs {group_shape_elements}")
        assert len(elements) == group_shape_elements


# working with Pages

@pytest.mark.parametrize("filename, page_name", [("test1.vsdx", "Page-1"), ("test2.vsdx", "Page-1")])
def test_get_page(filename: str, page_name: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        # confirm page name as expected
        assert page.name == page_name


@pytest.mark.parametrize(("filename"),
                         [("test1.vsdx"),
                          ])
def test_app_xml_page_names(filename: str):
    # test that page names in app.xml matches page names loaded
    with VisioFile(os.path.join(basedir, filename)) as vis:
        HeadingPairs = vis.app_xml.getroot().find(f'{ext_prop_namespace}HeadingPairs')
        i4 = HeadingPairs.find(f'.//{vt_namespace}i4')
        num_pages = int(i4.text)
        assert num_pages == len(vis.pages)

        # check page names from pages is same as page names from app.xml
        page_names = [p.name for p in vis.pages]
        TitlesOfParts = vis.app_xml.getroot().find(f'{ext_prop_namespace}TitlesOfParts')
        vector = TitlesOfParts.find(f'{vt_namespace}vector')
        app_xml_page_names = []
        for lpstr in vector.findall(f'{vt_namespace}lpstr'):
            page_name = lpstr.text
            app_xml_page_names.append(page_name)
        assert page_names == app_xml_page_names


@pytest.mark.parametrize(("filename", "page_index"),
                         [
                             ("test1.vsdx", 0),
                             ("test5_master.vsdx", 0),
                         ]
                         )
def test_remove_page_by_index(filename: str, page_index: int):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_remove_page.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page_count = len(vis.pages)
        vis.remove_page_by_index(page_index)
        vis.save_vsdx(out_file)

    # re-open file and confirm it has one less page
    with VisioFile(out_file) as vis:
        assert len(vis.pages) == page_count - 1


@pytest.mark.parametrize(("filename", "page_name"),
                         [
                             ("test1.vsdx", '1'),
                             ("test2.vsdx", '2'),
                             ("test3_house.vsdx", '1'),
                             ("test4_connectors.vsdx", '2'),
                             ("test5_master.vsdx", '1'),
                             ("test6_shape_properties.vsdx", '2'),
                             ("test7_with_connector.vsdx", '3'),
                             ("test8_simple_connector.vsdx", 'none'),
                         ])
def test_remove_page_by_page_index(filename: str, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_remove_page_by_page_index.vsdx')
    expected_page_names = []
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"Found {len(vis.pages)} pages, {sorted([p.name for p in vis.pages])}")
        for page in list(vis.pages):
            print(f"page.name='{page.name}' page.page_name='{page.page_name}' index:{page.index_num}")
            if page_name in page.name:
                print(f"Removing page index {page.index_num}")
                vis.remove_page_by_index(page.index_num)
            else:
                expected_page_names.append(page.name)

        vis.save_vsdx(out_file)

    print(f"expected names={expected_page_names}")
    # re-open file and confirm it contains all and only those not deleted
    with VisioFile(out_file) as vis:
        print(f"Found {len(vis.pages)} pages, {sorted([p.name for p in vis.pages])}")
        assert sorted(expected_page_names) == sorted([p.name for p in vis.pages])


@pytest.mark.parametrize(("filename", "page_name"),
                         [
                             ("test1.vsdx", 'Page-1'),
                             ("test2.vsdx", 'Page-2'),
                             ("test3_house.vsdx", 'Page-1'),
                             ("test4_connectors.vsdx", 'Page-2'),
                             ("test5_master.vsdx", 'Page 1'),
                             ("test6_shape_properties.vsdx", 'Page-2'),
                             ("test7_with_connector.vsdx", 'Page-3'),
                             ("test8_simple_connector.vsdx", 'none'),  # no match
                         ])
def test_remove_page_by_name(filename: str, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_remove_page_by_name.vsdx')
    expected_page_names = []
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"Found {len(vis.pages)} pages, {sorted([p.name for p in vis.pages])}")
        for page in list(vis.pages):
            print(f"page.name='{page.name}' page.page_name='{page.page_name}' index:{page.index_num}")
            if page_name != page.name:
                expected_page_names.append(page.name)
        vis.remove_page_by_name(page_name)
        vis.save_vsdx(out_file)

    print(f"expected names={expected_page_names}")
    # re-open file and confirm it contains all and only those not deleted
    with VisioFile(out_file) as vis:
        print(f"Found {len(vis.pages)} pages, {sorted([p.name for p in vis.pages])}")
        assert sorted(expected_page_names) == sorted([p.name for p in vis.pages])


@pytest.mark.parametrize(("filename", "remove_index"),
                         [("test1.vsdx", 0),
                          ("test1.vsdx", 1),
                          ("test2.vsdx", 0),
                          ])
def test_app_xml_page_names_after_remove_page(filename: str, remove_index: int):
    # test that page names in app.xml matches page names loaded
    with VisioFile(os.path.join(basedir, filename)) as vis:
        vis.remove_page_by_index(remove_index)

        HeadingPairs = vis.app_xml.getroot().find(f'{ext_prop_namespace}HeadingPairs')
        i4 = HeadingPairs.find(f'.//{vt_namespace}i4')
        num_pages = int(i4.text)
        assert num_pages == len(vis.pages)

        # check page names from pages is same as page names from app.xml
        page_names = [p.name for p in vis.pages]
        TitlesOfParts = vis.app_xml.getroot().find(f'{ext_prop_namespace}TitlesOfParts')
        vector = TitlesOfParts.find(f'{vt_namespace}vector')
        app_xml_page_names = []
        for lpstr in vector.findall(f'{vt_namespace}lpstr'):
            page_name = lpstr.text
            app_xml_page_names.append(page_name)
        assert page_names == app_xml_page_names


@pytest.mark.parametrize(('filename'),
                         [('test1.vsdx')])
def test_add_page(filename: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_add_page.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        number_pages = len(vis.pages)

        new_page = vis.add_page()
        assert new_page
        assert len(vis.pages) == number_pages + 1

        new_page_name = new_page.name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(new_page_name)
        assert page


@pytest.mark.parametrize(('filename', 'page_name'),
                         [
                            ('test1.vsdx', 'newname'),
                            ('test1.vsdx', 'Page-1'),
                         ])
def test_add_page_name(filename: str, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_add_page_name_{page_name}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        number_pages = len(vis.pages)

        new_page = vis.add_page(page_name)
        assert new_page
        assert len(vis.pages) == number_pages + 1

        new_page_name = new_page.name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(new_page_name)
        assert page


@pytest.mark.parametrize(('filename', 'index', 'page_name'),
                         [
                            ('test1.vsdx', 1, None),
                            ('test1.vsdx', 1, 'newname'),
                            ('test1.vsdx', 1, 'Page-1'),
                         ])
def test_add_page_at(filename: str, index: int, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_add_page_at_{page_name}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        number_pages = len(vis.pages)

        new_page = vis.add_page_at(index, page_name)
        assert new_page
        assert len(vis.pages) == number_pages + 1

        new_page_name = new_page.name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(new_page_name)
        assert page


@pytest.mark.parametrize(("filename", "new_page_name", "location"),
                         [("test1.vsdx", "new_page", 0),
                          ("test1.vsdx", "new_page", 1),
                          ("test2.vsdx", "new_page", 0),
                          ])
def test_app_xml_page_names_after_add_page(filename: str, new_page_name: str, location: int):
    # test that page names in app.xml matches page names loaded
    with VisioFile(os.path.join(basedir, filename)) as vis:
        if location is not None:
            vis.add_page(new_page_name)
        else:
            vis.add_page_at(location, new_page_name)

        HeadingPairs = vis.app_xml.getroot().find(f'{ext_prop_namespace}HeadingPairs')
        i4 = HeadingPairs.find(f'.//{vt_namespace}i4')
        num_pages = int(i4.text)
        assert num_pages == len(vis.pages)

        # check page names from pages is same as page names from app.xml
        page_names = [p.name for p in vis.pages]
        TitlesOfParts = vis.app_xml.getroot().find(f'{ext_prop_namespace}TitlesOfParts')
        vector = TitlesOfParts.find(f'{vt_namespace}vector')
        app_xml_page_names = []
        for lpstr in vector.findall(f'{vt_namespace}lpstr'):
            page_name = lpstr.text
            app_xml_page_names.append(page_name)
        assert page_names == app_xml_page_names


@pytest.mark.parametrize(('filename', 'index', 'page_name'),
                         [
                            ('test1.vsdx', -1, None),
                            ('test1.vsdx', 0, 'newname'),
                            ('test1.vsdx', 2, 'Page-1'),
                            ('test2.vsdx', 2, 'Page-1'),
                            ('test4_connectors.vsdx', 0, 'Page-1'),
                         ])
def test_copy_page(filename: str, index: int, page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_page_{page_name}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        number_pages = len(vis.pages)

        page = vis.pages[0]  # type: Page
        new_page = vis.copy_page(page, index=index, name=page_name)
        assert new_page
        assert len(vis.pages) == number_pages + 1

        new_page_name = new_page.name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(new_page_name)
        assert page


@pytest.mark.parametrize(('filename', 'page_index_to_copy', 'in_page_name', 'out_page_name'),
                         [
                             ('test1.vsdx', 0, None, 'Page-1-1'),
                             ('test1.vsdx', 0,  'newname', 'newname'),
                             ('test1.vsdx', 0, 'Page-1', 'Page-1-1'),
                             ('test1.vsdx', 1, None, 'Page-2-1'),
                             ('test1.vsdx', 1,  'newname', 'newname'),
                             ('test1.vsdx', 1, 'Page-1', 'Page-1-1'),
                         ])
def test_copy_page_naming(filename: str, page_index_to_copy:int, in_page_name: str, out_page_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_page_naming.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index_to_copy]  # type: Page
        new_page = vis.copy_page(page, name=in_page_name)
        print(f"in_page_name:{in_page_name} out_page_name:{out_page_name} actual:{new_page.name}")
        assert new_page.name == out_page_name  # check new page has expected name
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(out_page_name)
        assert page  # check that page name persists through file save and open


@pytest.mark.parametrize(('filename', 'page_index_to_copy', 'page_position', 'out_page_index'),
                         [
                             ('test1.vsdx', 0, PagePosition.LAST, 3),
                             ('test1.vsdx', 0, PagePosition.BEFORE, 0),
                             ('test1.vsdx', 0, PagePosition.AFTER, 1),
                             ('test1.vsdx', 1, PagePosition.LAST, 3),
                             ('test1.vsdx', 1, PagePosition.BEFORE, 1),
                             ('test1.vsdx', 1, PagePosition.AFTER, 2),
                         ])
def test_copy_page_positions(filename: str, page_index_to_copy:int, page_position: PagePosition, out_page_index: int):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_page_position.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[page_index_to_copy]  # type: Page
        new_page = vis.copy_page(page, index=page_position)
        index = vis.pages.index(new_page)
        print(f"page_index_to_copy:{page_index_to_copy} page_position:{page_position} out_page_index:{out_page_index} "
              f"actual:{index}")
        assert index == out_page_index  # check new page has expected index
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page_by_name(new_page.name)
        index = vis.pages.index(page)
        assert index == out_page_index  # check that page location persists through file save and open


# working with Shapes

@pytest.mark.parametrize(("filename", "shape_name"),
                         [("test1.vsdx", "Shape to copy"),
                          ("test4_connectors.vsdx", "Shape B")])
def test_vis_copy_shape(filename: str, shape_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_vis_copy_shape.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        # find and copy shape by name
        s = page.find_shape_by_text(shape_name)  # type: Shape
        assert s  # check shape found
        print(f"Found shape id:{s.ID}")
        max_id = page.max_id

        # note = this does add the shape, but prefer Shape.copy() as per next test which wraps this and returns Shape
        page.set_max_ids()
        new_shape = vis.copy_shape(shape=s.xml, page=page)

        assert new_shape  # check copy_shape returns xml

        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.attrib['ID']}")
        assert int(new_shape.attrib.get('ID')) > int(s.ID)
        assert int(new_shape.attrib.get('ID')) > max_id

        new_shape_id = new_shape.attrib['ID']
        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        s = page.find_shape_by_id(new_shape_id)
        assert s


@pytest.mark.parametrize(("filename", "shape_name"),
                         [("test1.vsdx", "Shape to copy"),
                          ("test2.vsdx", "Shape to copy")])
def test_copy_shape_other_page(filename: str, shape_name: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_copy_shape_other_page.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page =  vis.pages[0]  # type: Page
        page2 = vis.pages[1]  # type: Page
        page3 = vis.pages[2]  # type: Page
        # find and copy shape by name
        s = page.find_shape_by_text(shape_name)  # type: Shape
        assert s  # check shape found
        shape_text = s.text
        print(f"Found shape id:{s.ID}")

        new_shape = vis.copy_shape(shape=s.xml, page=page2)
        assert new_shape  # check copy_shape returns xml
        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.attrib['ID']}")
        page2_new_shape_id = new_shape.attrib['ID']

        new_shape = vis.copy_shape(shape=s.xml, page=page3)
        assert new_shape  # check copy_shape returns xml
        print(f"created new shape {type(new_shape)} {new_shape} {new_shape.attrib['ID']}")
        page3_new_shape_id = new_shape.attrib['ID']

        vis.save_vsdx(out_file)

    # re-open saved file and check it is changed as expected
    with VisioFile(out_file) as vis:
        page2 = vis.pages[1]
        s = page2.find_shape_by_id(page2_new_shape_id)
        assert s
        assert s.text == shape_text
        page3 = vis.pages[2]
        s = page3.find_shape_by_id(page3_new_shape_id)
        assert s
        assert s.text == shape_text



import pprint
import pytest
import os

from vsdx import VisioFile
from vsdx.vsdxdiff import VisioFileDiff


# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))


@pytest.mark.parametrize(("filename_a", "filename_b"),
                         [("test1.vsdx", "test2.vsdx" ),
                          ("test1.vsdx", "test4_connectors.vsdx" )])
def test_create_visiodiff(filename_a: str, filename_b: str):
    filepath_a = os.path.join(basedir, filename_a)
    filepath_b = os.path.join(basedir, filename_b)
    print(basedir)
    print(filepath_a)
    fd = VisioFileDiff(filepath_a, filepath_b)
    print(f"fd={fd}")
    print(f"Added in {filename_b} {fd.added_members()}")
    print(f"Removed in {filename_b} {fd.removed_members()}")
    for m in fd.common_members():
        print(f"\n\n{m}")
        pprint.pprint(fd.diffs.get(m))


# next test, open file, set text of shape, save as - then compare the two
@pytest.mark.skip
@pytest.mark.parametrize(("filename_a", "filename_b"),
                         [("test1.vsdx", "test1_outfile.vsdx" ),
                          ("test2.vsdx", "test2_outfile.vsdx" ),
                          ])
def test_visiodiff_before_after(filename_a: str, filename_b: str):
    filepath_a = os.path.join(basedir, filename_a)
    filepath_b = os.path.join(basedir, 'out', filename_b)
    with VisioFile(filepath_a) as vis:
        print(f"saving as {filepath_b}")
        vis.save_vsdx(filepath_b)

    fd = VisioFileDiff(filepath_a, filepath_b)

    print(f"fd={fd}")
    print(f"Added in {filename_b} {fd.added_members()}")
    print(f"Removed in {filename_b} {fd.removed_members()}")
    for m in fd.common_members():
        diff = fd.diffs.get(m)
        print(f"\n\n{m} {type(diff)} len:{len(diff) if diff else 0}")
        if diff:
            num = 0
            for l in diff:
                num += 1
                print(f"{num} {l}")


@pytest.mark.skip
@pytest.mark.parametrize(("filename_a", "filename_b"),
                         [
                             ("test4_connectors_out.vsdx", "test4_connectors_added.vsdx" ),
                          ])
def test_visiodiff_two_files(filename_a: str, filename_b: str):
    filepath_a = os.path.join(basedir, filename_a)
    filepath_b = os.path.join(basedir, filename_b)

    fd = VisioFileDiff(filepath_a, filepath_b)

    print(f"fd={fd}")
    print(f"Added in {filename_b} {fd.added_members()}")
    print(f"Removed in {filename_b} {fd.removed_members()}")
    for m in fd.common_members():
        diff = fd.diffs.get(m)
        print(f"\n\n{m} {type(diff)} len:{len(diff) if diff else 0}")
        if diff:
            num = 0
            for l in diff:
                num += 1
                print(f"{num} {l}")


import os
from .vsdxfile import VisioFile


class Media:
    straight_connector_text = 'STRAIGHT_CONNECTOR'
    curved_connector_text = 'CURVED_CONNECTOR'
    rectangle_text = "RECTANGLE"
    circle_text = "CIRCLE"

    def __init__(self):
        basedir = str(os.path.relpath(__file__))
        file_path = os.sep.join(basedir.split(os.sep)[:-1])
        file_path = os.path.join(file_path, 'media', 'media.vsdx')
        self._media_vsdx = VisioFile(file_path)

    @property
    def rels_xml(self):
        return self._media_vsdx.pages[0].rels_xml

    @property
    def straight_connector(self):
        return self._media_vsdx.pages[0].find_shape_by_text(Media.straight_connector_text)

    @property
    def curved_connector(self):
        return self._media_vsdx.pages[0].find_shape_by_text(Media.straight_connector_text)

    @property
    def rectangle(self):
        return self._media_vsdx.pages[0].find_shape_by_text(Media.rectangle_text)

    @property
    def circle(self):
        return self._media_vsdx.pages[0].find_shape_by_text(Media.circle_text)



import pytest
import os
from datetime import datetime

from vsdx import Page  # for typing
from vsdx import Shape  # for typing
from vsdx import VisioFile

# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))

@pytest.mark.parametrize(("filename", "context"),
                         [("test_jinja.vsdx", {"date": datetime.now(), "scenario": "Scenario One", "x": 2, "y": 2}),
                          ("test_jinja.vsdx", {"date": datetime.now(), "scenario": "Scenario Two", "x": 2, "y": 2}),
                          ("test_jinja.vsdx", {"date": datetime.now(), "scenario": "Scenario Three", "x": 2, "y": 2}),
                          ])
def test_basic_jinja(filename: str, context: dict):

    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_basic_jinja.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]

        # each key string in context dict will be replaced with value, record against shape Ids for validation
        shape_id_values = dict()
        for k, v in context.items():
            shape_id = page.find_shape_by_text(k).ID
            shape_id_values[shape_id] = v
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and validate each shape id has expected text
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        for shape_id, text in shape_id_values.items():
            if type(text) is str:
                print(f"Testing that shape {shape_id} has text '{text}' in: {page.find_shape_by_id(shape_id).text}")
                assert str(text) in page.find_shape_by_id(shape_id).text


@pytest.mark.parametrize(("filename", "context", "shape_count"),
                         [("test_jinja.vsdx", {"date": datetime.now(), "scenario": "One", "x": 0, "y": 2}, 1),
                          ("test_jinja.vsdx", {"date": datetime.now(), "scenario": "Two", "x": 5, "y": 2}, 2),
                          ("test_jinja.vsdx", {"date": datetime.now(), "scenario": "Three", "x": 20, "y": 2}, 3),
                          ])
def test_jinja_if(filename: str, context: dict, shape_count: int):

    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_jinja_if_{context["scenario"]}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and validate each shape id has expected text
    with VisioFile(out_file) as vis:
        page = vis.pages[1]  # second page has the shapes with if statements
        count = len(page.child_shapes)
        print(f"expected {shape_count} and found {count}")
        assert count == shape_count


@pytest.mark.parametrize(("filename", "context"),
                         [("test_jinja.vsdx", {"x": 2, "y": 2}),
                          ("test_jinja.vsdx", {"x": 3, "y": 4}),
                          ("test_jinja.vsdx", {"x": 12, "y": 9}),
                          ])
def test_jinja_calc(filename: str, context: dict):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_jinja_calc.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and validate each shape id has expected text
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        # check a shape exists with product of x and y values
        x_y = str(context['x'] * context['y'])
        assert page.find_shape_by_text(x_y)


@pytest.mark.parametrize(("filename", "context"),
                         [("test_jinja_loop.vsdx", {"date": datetime.now(), "scenario": "Scenario One", "test_list":[1,2,3]}),
                          ("test_jinja_loop.vsdx", {"date": datetime.now(), "scenario": "Scenario Two", "test_list":["One", "Two","Three"]}),
                          ("test_jinja_loop.vsdx", {"date": datetime.now(), "scenario": "Scenario Three", "test_list":[1,2,3,4,5,6]}),
                          ])
def test_basic_jinja_loop(filename: str, context: dict):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_basic_jinja.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]

        # each key string in context dict will be replaced with value, record against shape Ids for validation
        shape_id_values = dict()
        for k, v in context.items():
            shape_id = page.find_shape_by_text(k).ID
            shape_id_values[shape_id] = v
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and validate each shape id has expected text, and that a shape exists with each loop value
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        for shape_id, text in shape_id_values.items():
            if type(text) is str:
                print(f"Testing that shape {shape_id} has text '{text}' in: {page.find_shape_by_id(shape_id).text}")
                assert str(text) in page.find_shape_by_id(shape_id).text
            if type(text) is list:
                for item in text:
                    print(f"Testing that shape with text '{item}' exists")
                    assert page.find_shape_by_text(str(item))


@pytest.mark.parametrize(("filename", "context"),
                         [("test_jinja_inner_loop.vsdx", {"test_list":[[1, 2, 3], [1, 2, 3], [1, 2, 3]]}),
                          ("test_jinja_inner_loop.vsdx", {"test_list":["One", "Two","Three"]}),
                          ])
def test_jinja_inner_loop(filename: str, context: dict):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_jinja_inner_loop.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and validate each shape id has expected text, and that a shape exists with each loop value
    test_list = context["test_list"]
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        for o in test_list:
            for p in o:
                print(f"p={p}")
                s = page.find_shape_by_text(str(p))
                assert s


@pytest.mark.parametrize(("filename", "out_name", "context"),
                         [("test_jinja_loop_showif.vsdx", "1234", {"test_list": [1, 2, 3, 4]}),
                          ("test_jinja_loop_showif.vsdx", "3456", {"test_list": [3, 4, 5, 6]}),
                          ])
def test_jinja_loop_showif(filename: str, out_name: str,  context: dict):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}{out_name}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        for o in context['test_list']:
            s = page.find_shape_by_text(f"In this instance, o={o}")
            # test that {% show if o > 2 %} has worked
            print(f"for {o} found {s}")
            if o > 2:
                assert s  # check that shape is found (not removed by showif)
            else:
                assert not s  # check that shape is not found (has been removed by showif)


@pytest.mark.parametrize(
    ("filename", "context", "shape_id", "expected_x", "expected_text"),
    [("test_jinja_self_refs.vsdx", {"n": 1}, "1", 2.0, "This text should remain  and x should be 2.0\n"),
     ("test_jinja_self_refs.vsdx", {"n": 2}, "2", 4.0, "This shape sets x to n * 2\n"),
     ("test_jinja_self_refs.vsdx", {"n": 1}, "3", 1.0, "This shape sets x to 1 if n else 2\n"),
     ("test_jinja_self_refs.vsdx", {"n": 0}, "3", 2.0, "This shape sets x to 1 if n else 2\n"),
     ])
def test_jinja_self_refs(filename: str, context: dict, shape_id, expected_x, expected_text):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_jinja_self_refs.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        # there should be one shape on page 0
        shape = page.find_shape_by_id(shape_id)  # type: Shape
        print(f"DEBUG: ID={shape.ID} shape.text={shape.text}")
        print(f"DEBUG: ID={shape.ID} shape.x={shape.x}")
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and check shape has moved
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        shape = page.find_shape_by_id(shape_id)  # type: Shape
        print(f"DEBUG: ID={shape.ID} shape.text='{shape.text}' expected='{expected_text}'")
        print(f"DEBUG: ID={shape.ID} shape.x={shape.x}")
        assert shape.x == expected_x
        assert shape.text == expected_text


@pytest.mark.parametrize(
    ("filename", "context", "shape_id", "expected_y", "expected_text"),
    [("test_jinja_self_refs.vsdx", {"n": 2}, "4", 10.12368731806121, "This shape should move down by 1.0\n"),
     ("test_jinja_self_refs.vsdx", {"n": 1}, "5", 8.726049539918966, "This shape should move down by n\n"),
     ("test_jinja_self_refs.vsdx", {"n": 2}, "5", 7.726049539918966, "This shape should move down by n\n"),
     ])
def test_jinja_self_ref_calculations(filename: str, context: dict, shape_id, expected_y, expected_text):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_jinja_self_ref_calcs.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.pages[0]  # type: Page
        # there should be one shape on page 0
        shape = page.find_shape_by_id(shape_id)  # type: Shape
        if shape:
            print(f"DEBUG: ID={shape.ID} shape.text={shape.text}")
            print(f"DEBUG: ID={shape.ID} shape.y={shape.y}")
        vis.jinja_render_vsdx(context=context)
        vis.save_vsdx(out_file)

    # open file and check shape has moved
    with VisioFile(out_file) as vis:
        page = vis.pages[0]
        shape = page.find_shape_by_id(shape_id)  # type: Shape
        print(f"DEBUG: ID={shape.ID} shape.text='{shape.text}' expected='{expected_text}'")
        print(f"DEBUG: ID={shape.ID} shape.x={shape.y}")
        assert shape.y == expected_y
        assert shape.text == expected_text


@pytest.mark.parametrize(
    ("filename", "context", "expected_page_count", "expected_page_names"),
    [("test_jinja_page_showif.vsdx", {"show": True}, 2, ['Normal Page', 'Page2']),
     ("test_jinja_page_showif.vsdx", {"show": 1}, 2, ['Normal Page', 'Page2']),
     ("test_jinja_page_showif.vsdx", {"show": "true"}, 2, ['Normal Page', 'Page2']),
     ("test_jinja_page_showif.vsdx", {"show": 4.0}, 2, ['Normal Page', 'Page2']),
     ("test_jinja_page_showif.vsdx", {"show": False}, 2, ['Normal Page', 'Page3']),
     ("test_jinja_page_showif.vsdx", {"show": []}, 2, ['Normal Page', 'Page3']),
     ("test_jinja_page_showif.vsdx", {"show": {}}, 2, ['Normal Page', 'Page3']),
     ])
def test_jinja_page_showif(filename: str, context: dict, expected_page_count, expected_page_names):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_show_{context["show"]}.vsdx')
    with VisioFile(os.path.join(basedir, filename)) as vis:
        print(f"len(vis.pages)={len(vis.pages)} context={context}")
        print("BEFORE", [(p.index_num, p.name) for p in vis.pages])
        vis.jinja_render_vsdx(context=context)
        print("AFTER", [(p.index_num, p.name) for p in vis.pages])
        vis.save_vsdx(out_file)

    # open file and check shape has moved
    with VisioFile(out_file) as vis:
        page_names = []
        for p in vis.pages: # type: Page
            print(f"page:{p.name}")
            page_names.append(p.name)
        assert len(vis.pages) == expected_page_count
        assert page_names == expected_page_names



import pytest
import os

import vsdx
from vsdx import Page  # for typing
from vsdx import Shape  # for typing
from vsdx import VisioFile

# code to get basedir of this test file in either linux/windows
basedir = os.path.dirname(os.path.relpath(__file__))


@pytest.mark.parametrize(("filename", "expected_length"),
                         [('test5_master.vsdx', 1),
                          ('test3_house.vsdx', 1),])
def test_load_master_file(filename: str, expected_length: int):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        assert len(vis.master_pages) == expected_length


@pytest.mark.parametrize(("filename"),
                         [('test3_house.vsdx'),
                          ('test_master_multiple_child_shapes.vsdx'),])
def test_master_unique_id(filename: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        # note: master base id does not exist in test5_master.vsdx a file created by LucidChart
        for m in vis.master_pages:
            # check master base id exists
            assert m.master_unique_id is not None


@pytest.mark.parametrize(("filename"),
                         [('test3_house.vsdx'),
                          ('test_master_multiple_child_shapes.vsdx'),])
def test_master_base_id(filename: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        # note: master base id does not exist in test5_master.vsdx a file created by LucidChart
        for m in vis.master_pages:
            # check master base id exists
            assert m.master_base_id is not None


@pytest.mark.parametrize(("filename", "shape_text"),
                         [('test5_master.vsdx', "Shape B")])
def test_find_master_shape(filename: str, shape_text: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        master_page =  vis.master_pages[0]  # type: Page
        s = master_page.find_shape_by_text(shape_text)
        assert s


@pytest.mark.parametrize(("filename"),
                         [('test5_master.vsdx')])
def test_master_inheritance(filename: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        master_page = vis.master_pages[0]  # type: Page
        shape_a = page.find_shape_by_text('Shape A')  # type: Shape
        shape_b = page.find_shape_by_text('Shape B')  # type: Shape

        assert shape_a
        assert shape_b

        # test inheritance of cell values
        assert shape_a.width == 1
        assert shape_a.height == 0.75

        # test inheritance for subshapes
        sub_shape_a = shape_a.child_shapes[0]
        assert sub_shape_a.cell_value('LineWeight') == '0.01875'


@pytest.mark.parametrize(("filename"),
                         [('test5_master.vsdx')])
def test_set_master_child_property(filename: str):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_set_master_child_property.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)
        # find shape with master and set a child property
        sub_shape_b = page.find_shape_by_text('Shape B').child_shapes[0]

        sub_shape_b.line_weight = 0.5

        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page(0)

        sub_shape_b = page.find_shape_by_text('Shape B').child_shapes[0]
        assert sub_shape_b.master_shape  # shape has a master
        assert sub_shape_b.line_weight == 0.5  # child shape has value set


@pytest.mark.parametrize(("filename, weight"),
                         [('test5_master.vsdx', 0.5)])
def test_master_property_change_is_inherited(filename: str, weight):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_master_property_change_is_inherited.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)

        sub_shape_a = page.find_shape_by_text('Shape A').child_shapes[0]
        master = sub_shape_a.master_shape
        master.line_weight = weight

        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page(0)

        sub_shape_a = page.find_shape_by_text('Shape A').child_shapes[0]
        sub_shape_b = page.find_shape_by_text('Shape B').child_shapes[0]

        # check that both sub_shape values have changed based on master
        assert sub_shape_a.line_weight == weight
        assert sub_shape_b.line_weight == weight


@pytest.mark.parametrize(("filename, weight"),
                         [('test5_master.vsdx', 0.1)])
def test_child_property_change_is_not_inherited(filename: str, weight):
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_child_property_change_is_not_inherited.vsdx')

    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)

        shape_a = page.find_shape_by_text('Shape A')
        sub_shape_a = shape_a.child_shapes[0]
        master = sub_shape_a.master_shape
        original_master_weight = master.line_weight
        # increment child property to ensure child and master are not the same
        line_weight = master.line_weight + weight
        sub_shape_a.line_weight = line_weight  # change the child, not the master shape

        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page(0)

        sub_shape_a = page.find_shape_by_text('Shape A').child_shapes[0]
        master = sub_shape_a.master_shape

        # check that child shape has changed to expected value
        assert sub_shape_a.line_weight == line_weight
        # check that master property has not changed
        assert master.line_weight == original_master_weight


@pytest.mark.parametrize(("filename", "shape_text"),
                         [("test_master.vsdx", "Page Shape"),
                          ("test_master.vsdx", "Master Shape A"),
                          ("test_master.vsdx", "Master Shape B"),
                          ("test_master.vsdx", "Master B with updated text"),
                          ])
def test_master_find_shapes(filename: str, shape_text: str):
    # Check that shape with text can be found - whether in page, master or overridden in master
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_text(shape_text)
        assert shape  # shape found


@pytest.mark.parametrize(("filename", "shape_text", "has_master"),
                         [("test1.vsdx", "Shape to remove", False),  # no master file
                          ("test_master.vsdx", "Page Shape", False),  # shape with no master
                          ("test_master.vsdx", "Master Shape A", True),  # shape with master
                          ("test_master.vsdx", "Master Shape B", True),
                          ("test_master.vsdx", "Master B with updated text", True),
                          ])
def test_shape_has_master(filename: str, shape_text: str, has_master: bool):
    # Check that shape.master_shape returns a shape or None as expected
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_text(shape_text)
        assert shape  # shape found
        if has_master:
            assert shape.master_shape  # shape has a master
            assert shape.master_shape.is_master_shape  # master is a valid master shape
        else:
            assert not shape.master_shape  # shape does not have a master


@pytest.mark.parametrize(("filename", "shape_text", "has_master", "inherits_text"),
                         [("test_master.vsdx", "Page Shape", False, False),
                          ("test_master.vsdx", "Master Shape A", True, True),
                          ("test_master.vsdx", "Master Shape B", True, True),
                          ("test_master.vsdx", "Master B with updated text", True, False), ])
def test_master_check_text_inheritance(filename: str, shape_text: str, has_master: bool, inherits_text: bool):
    # Check that shape with text can be found and that it has a master (or not) and inherits text (or not)
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)  # type: Page
        shape = page.find_shape_by_text(shape_text)

        if shape.master_shape:
            assert has_master
            assert (shape.text == shape.master_shape.text) == inherits_text
        else:
            assert not has_master


@pytest.mark.parametrize(("filename", "master_shape_text", "number_shapes"),
                         [('test_master.vsdx', 'Master Shape A', 1),
                          ('test_master.vsdx', 'Master Shape B', 2),
                          ])
def test_find_shapes_by_master_id(filename: str, master_shape_text: str, number_shapes: int):
    # test that a shapes master has 'number_shapes' shapes that inherit from it
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)
        shape = page.find_shape_by_text(master_shape_text)  # get shape which has a master
        # print(f"master_shape: {master_shape} {master_shape_.master_shape_ID}")
        shapes = page.find_shapes_with_same_master(shape)
        print(f"shapes: {shapes}")
        assert len(shapes) == number_shapes


@pytest.mark.parametrize(("filename", "master_shape_search_text", "shape_search_text", "expect_inherit"),
                         [('test_master.vsdx', 'Master Shape A', 'Master Shape A', True),
                          ('test_master.vsdx', 'Master Shape A', 'Master B with updated text', False),
                          ('test_master.vsdx', 'Master B with updated text', 'Master B with updated text', False),
                          ('test_master.vsdx', 'Master Shape B', 'Master Shape B', True),
                          ('test_master.vsdx', 'Master Shape A', 'Page Shape', False),
                          ])
def test_master_inheritance_master_shape_set_text(filename: str, master_shape_search_text: str, shape_search_text: str,
                                                  expect_inherit: bool):
    # test that when a master shape text is updated, only shapes that inherit that master shapes text are affected
    out_file = os.path.join(basedir, 'out', f'{filename[:-5]}_test_master_inheritance_master_shape_set_text.vsdx')
    master_text = "Updated Master Shape Text"
    with VisioFile(os.path.join(basedir, filename)) as vis:
        page = vis.get_page(0)

        # get master shape of Shape with search text
        master_shape = page.find_shapes_by_text(master_shape_search_text)[0].master_shape
        shape = page.find_shape_by_text(shape_search_text)
        shape_id = shape.ID
        shape_text = shape.text
        master_shape.text = master_text

        if expect_inherit:
            assert shape.text == master_text  # inherited from master change
        else:
            assert shape.text == shape_text  # unchanged

        vis.save_vsdx(out_file)

    with VisioFile(out_file) as vis:
        page = vis.get_page(0)

        # get same shape and confirm inheritance works as expected
        shape = page.find_shape_by_id(shape_id)
        if expect_inherit:
            assert shape.text == master_text  # inherited from master change
        else:
            assert shape.text == shape_text  # unchanged


@pytest.mark.parametrize(
    "filename, shape_id, expected_text",
    [
        ("test_master_multiple_child_shapes.vsdx", '3', 'AWS Step Functions workflow \n'),
        ("test_master.vsdx", '4', 'Master Shape A\n'),
        ("test_master.vsdx", '10', 'Master Shape B\n'),
        ("test_master.vsdx", '11', 'Master B with updated text\n'),
    ])
def test_get_text_from_master_shape(filename: str, shape_id: str, expected_text: str):
    with VisioFile(os.path.join(basedir, filename)) as vis:
        # print out ID and text for all shapes in first page
        for s in vis.pages[0].all_shapes:
            print(f"ID={s.ID} text='{s.text}'")

        child_shape = vis.pages[0].find_shape_by_id(shape_id)
        assert child_shape.text == expected_text



namespace = "{http://schemas.microsoft.com/office/visio/2012/main}"  # visio file name space
ext_prop_namespace = '{http://schemas.openxmlformats.org/officeDocument/2006/extended-properties}'
vt_namespace = '{http://schemas.openxmlformats.org/officeDocument/2006/docPropsVTypes}'
r_namespace = '{http://schemas.openxmlformats.org/officeDocument/2006/relationships}'
document_rels_namespace = "{http://schemas.openxmlformats.org/package/2006/relationships}"
cont_types_namespace = '{http://schemas.openxmlformats.org/package/2006/content-types}'

# Ref: https://docs.microsoft.com/en-us/office/client-developer/visio/visio-file-format-reference

import xml.etree.ElementTree as ET
from xml.etree.ElementTree import Element
import xml.dom.minidom as minidom   # minidom used for prettyprint


def pretty_print_element(xml: Element) -> str:
    if type(xml) is Element:
        return minidom.parseString(ET.tostring(xml)).toprettyxml()
    elif type(xml) is ET.ElementTree:
        return minidom.parseString(ET.tostring(xml.getroot())).toprettyxml()
    else:
        return f"Not an Element. type={type(xml)}"


__version__ = "0.5.20"
from .shapes import Cell
from .connectors import Connect
from .shapes import DataProperty
from .pages import Page
from .pages import PagePosition
from .shapes import Shape
from .formulae import calc_value
from .vsdxfile import VisioFile
from .media import Media
from .geometry import Geometry, GeometryRow, GeometryCell
